/*
SaneGL - a small library to bring back sanity to WebGL
Copyright (C) 2011  John Tsiombikas <nuclear@member.fsf.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
const GL_MODELVIEW = 0;
const GL_PROJECTION = 1;
const GL_TEXTURE = 2;

const GL_POINTS = 1;
const GL_LINES = 2;
const GL_TRIANGLES = 3;
const GL_QUADS = 4;

var gl_ident_val = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1];
var gl_mmode = GL_MODELVIEW;
var gl_mat = new Array(
		[new Float32Array(gl_ident_val)],
		[new Float32Array(gl_ident_val)],
		[new Float32Array(gl_ident_val)]);

const GL_MAX_VERTS = 512;

var gl_prim = 0;
var gl_vbuf = null, gl_nbuf = null, gl_cbuf = null, gl_tbuf = null;
var gl_vertex, gl_normal, gl_color, gl_texcoord;
var gl_nverts, gl_vert_calls;
var gl_curr_normal = new Float32Array(3);
var gl_curr_color = new Float32Array(4);
var gl_curr_texcoord = new Float32Array(2);
var gl_vloc, gl_nloc, gl_cloc, gl_tloc;
var gl_use_normal, gl_use_color, gl_use_texcoord;

function glMatrixMode(mode)
{
	gl_mmode = mode;
}

function glPushMatrix()
{
	var mtop = gl_mat[gl_mmode].length - 1;

	gl_mat[gl_mmode].push(new Float32Array(16));
	m4_copy(gl_mat[gl_mmode][mtop + 1], gl_mat[gl_mmode][mtop]);
}

function glPopMatrix()
{
	if(gl_mat[gl_mmode].length <= 1) {
		alert('glPopMatrix underflow');
	}
	gl_mat[gl_mmode].pop();
}

function glLoadIdentity()
{
	var mtop = gl_mat[gl_mmode].length - 1;
	m4_identity(gl_mat[gl_mmode][mtop]);
}

function glLoadMatrixf(mat)
{
	var mtop = gl_mat[gl_mmode].length - 1;
	m4_copy(gl_mat[gl_mmode][mtop], mat);
}

function glMultMatrixf(mat)
{
	var mtop = gl_mat[gl_mmode].length - 1;
	m4_mul(gl_mat[gl_mmode][mtop], gl_mat[gl_mmode][mtop], mat);
}

function glTranslatef(x, y, z)
{
	var xform = new Float32Array(16);
	m4_translation(xform, x, y, z);
	glMultMatrixf(xform);
}

function glRotatef(angle, x, y, z)
{
	var angle_rads = Math.PI * angle / 180.0;
	var xform = new Float32Array(16);
	m4_rotation(xform, angle_rads, x, y, z);
	glMultMatrixf(xform);
}

function glScalef(x, y, z)
{
	var xform = new Float32Array(16);
	m4_scale(xform, x, y, z);
	glMultMatrixf(xform);
}

function glOrtho(left, right, bottom, top, near, far)
{
	var dx = right - left;
	var dy = top - bottom;
	var dz = far - near;

	var tx = -(right + left) / dx;
	var ty = -(top + bottom) / dy;
	var tz = -(far + near) / dz;

	var sx = 2.0 / dx;
	var sy = 2.0 / dy;
	var sz = -2.0 / dz;

	var xform = new Float32Array(16);
	xform[0] = sx;
	xform[5] = sy;
	xform[10] = sz;
	xform[12] = tx;
	xform[13] = ty;
	xform[14] = tz;
	xform[15] = 1.0;
	xform[1] = xform[2] = xform[3] = xform[4] = xform[6] = xform[7] = xform[8] = xform[9] = 0.0;

	glMultMatrixf(xform);
}

function glFrustum(left, right, bottom, top, near, far)
{
	var dx = right - left;
	var dy = top - bottom;
	var dz = far - near;

	var a = (right + left) / dx;
	var b = (top + bottom) / dy;
	var c = -(far + near) / dz;
	var d = -2.0 * far * near / dz;

	var xform = new Float32Array(16);
	xform[0] = 2.0 * near / dx;
	xform[5] = 2.0 * near / dy;
	xform[8] = a;
	xform[9] = b;
	xform[10] = c;
	xform[11] = -1.0;
	xform[14] = d;
	xform[1] = xform[2] = xform[3] = xform[4] = xform[6] = xform[7] = xform[12] = xform[13] = xform[15] = 0.0;

	glMultMatrixf(xform);
}

function gluPerspective(vfov, aspect, near, far)
{
	var x = near * Math.tan(vfov / 2.0);
	glFrustum(-aspect * x, aspect * x, -x, x, near, far);
}

function gl_apply_xform(prog)
{
	var mvtop = gl_mat[GL_MODELVIEW].length - 1;
	var ptop = gl_mat[GL_PROJECTION].length - 1;
	var ttop = gl_mat[GL_TEXTURE].length - 1;

	var loc = gl.getUniformLocation(prog, "mvmat");
	if(loc != -1) {
		gl.uniformMatrix4fv(loc, false, gl_mat[GL_MODELVIEW][mvtop]);
	}

	loc = gl.getUniformLocation(prog, "projmat");
	if(loc != -1) {
		gl.uniformMatrix4fv(loc, false, gl_mat[GL_PROJECTION][ptop]);
	}

	loc = gl.getUniformLocation(prog, "texmat");
	if(loc != -1) {
		gl.uniformMatrix4fv(loc, false, gl_mat[GL_TEXTURE][ttop]);
	}

	loc = gl.getUniformLocation(prog, "normmat");
	if(loc != -1) {
		var normmat = new Float32Array(16);
		m4_copy(normmat, gl_mat[GL_MODELVIEW][mvtop]);
		normmat[3] = normmat[7] = normmat[11] = normmat[12] = normmat[13] = normmat[14] = 0.0;
		normmat[15] = 1.0;
		gl.uniformMatrix4fv(loc, false, normmat);
	}
}

function glBegin(prim)
{
	gl_vertex = new Float32Array(GL_MAX_VERTS * 4);
	gl_normal = new Float32Array(GL_MAX_VERTS * 3);
	gl_color = new Float32Array(GL_MAX_VERTS * 4);
	gl_texcoord = new Float32Array(GL_MAX_VERTS * 4);

	if(gl_vbuf == null) {
		gl_vbuf = gl.createBuffer();
		gl.bindBuffer(gl.ARRAY_BUFFER, gl_vbuf);
		gl.bufferData(gl.ARRAY_BUFFER, GL_MAX_VERTS * 4 * Float32Array.BYTES_PER_ELEMENT, gl.STREAM_DRAW);
		logmsg("vertex buffer size: " + gl.getBufferParameter(gl.ARRAY_BUFFER, gl.BUFFER_SIZE) + "\n");

		gl_nbuf = gl.createBuffer();
		gl.bindBuffer(gl.ARRAY_BUFFER, gl_nbuf);
		gl.bufferData(gl.ARRAY_BUFFER, GL_MAX_VERTS * 3 * Float32Array.BYTES_PER_ELEMENT, gl.STREAM_DRAW);
		logmsg("normal buffer size: " + gl.getBufferParameter(gl.ARRAY_BUFFER, gl.BUFFER_SIZE) + "\n");

		gl_cbuf = gl.createBuffer();
		gl.bindBuffer(gl.ARRAY_BUFFER, gl_cbuf);
		gl.bufferData(gl.ARRAY_BUFFER, GL_MAX_VERTS * 4 * Float32Array.BYTES_PER_ELEMENT, gl.STREAM_DRAW);
		logmsg("color buffer size: " + gl.getBufferParameter(gl.ARRAY_BUFFER, gl.BUFFER_SIZE) + "\n");

		gl_tbuf = gl.createBuffer();
		gl.bindBuffer(gl.ARRAY_BUFFER, gl_tbuf);
		gl.bufferData(gl.ARRAY_BUFFER, GL_MAX_VERTS * 4 * Float32Array.BYTES_PER_ELEMENT, gl.STREAM_DRAW);
		logmsg("texcoord buffer size: " + gl.getBufferParameter(gl.ARRAY_BUFFER, gl.BUFFER_SIZE) + "\n");
	}

	gl_prim = prim;
	gl_nverts = gl_vert_calls = 0;

	gl_prog = gl.getParameter(gl.CURRENT_PROGRAM);
	if(gl_prog != null) {
		gl_apply_xform(gl_prog);

		gl_vloc = gl.getAttribLocation(gl_prog, "attr_vertex");
		gl_nloc = gl.getAttribLocation(gl_prog, "attr_normal");
		gl_cloc = gl.getAttribLocation(gl_prog, "attr_color");
		gl_tloc = gl.getAttribLocation(gl_prog, "attr_texcoord");
	} else {
		logmsg("no program currently bound\n");
		gl_vloc = gl_nloc = gl_cloc = gl_tloc = -1;
	}

	if(gl_vloc == -1) {
		logmsg("attr_vertex not found in currently bound program: " + prog + "\n");
	}

	gl_use_normal = gl_nloc != -1;
	gl_use_color = gl_cloc != -1;
	gl_use_texcoord = gl_tloc != -1;
}

function glEnd()
{
	if(gl_nverts > 0) {
		gl_draw_immediate();
	}
}

function gl_draw_immediate()
{
	var gles_prim;

	if(gl_vloc == -1) {
		return;
	}

	switch(gl_prim) {
	case GL_POINTS:
		gles_prim = gl.POINTS;
		break;
	case GL_LINES:
		gles_prim = gl.LINES;
		break;
	case GL_TRIANGLES:
	case GL_QUADS:
		gles_prim = gl.TRIANGLES;
		break;
	default:
	}

	gl.bindBuffer(gl.ARRAY_BUFFER, gl_vbuf);
	gl.bufferSubData(gl.ARRAY_BUFFER, 0, gl_vertex);
	gl.vertexAttribPointer(gl_vloc, 4, gl.FLOAT, false, 0, 0);
	gl.enableVertexAttribArray(gl_vloc);

	if(gl_use_normal) {
		gl.bindBuffer(gl.ARRAY_BUFFER, gl_nbuf);
		gl.bufferSubData(gl.ARRAY_BUFFER, 0, gl_normal);
		gl.vertexAttribPointer(gl_nloc, 3, gl.FLOAT, false, 0, 0);
		gl.enableVertexAttribArray(gl_nloc);
	}

	if(gl_use_color) {
		gl.bindBuffer(gl.ARRAY_BUFFER, gl_cbuf);
		gl.bufferSubData(gl.ARRAY_BUFFER, 0, gl_color);
		gl.vertexAttribPointer(gl_cloc, 4, gl.FLOAT, true, 0, 0);
		gl.enableVertexAttribArray(gl_cloc);
	}

	if(gl_use_texcoord) {
		gl.bindBuffer(gl.ARRAY_BUFFER, gl_tbuf);
		gl.bufferSubData(gl.ARRAY_BUFFER, 0, gl_texcoord);
		gl.vertexAttribPointer(gl_tloc, 4, gl.FLOAT, false, 0, 0);
		gl.enableVertexAttribArray(gl_tloc);
	}

	gl.drawArrays(gles_prim, 0, gl_nverts);

	gl.disableVertexAttribArray(gl_vloc);
	if(gl_use_normal) {
		gl.disableVertexAttribArray(gl_nloc);
	}
	if(gl_use_color) {
		gl.disableVertexAttribArray(gl_cloc);
	}
	if(gl_use_texcoord) {
		gl.disableVertexAttribArray(gl_tloc);
	}
}


function glVertex2f(x, y)
{
	glVertex3f(x, y, 0.0);
}

function glVertex3f(x, y, z)
{
	var i = 0;
	var vidx = gl_nverts * 4;
	var nidx = gl_nverts * 3;

	if(gl_prim == GL_QUADS && gl_vert_calls % 4 == 3) {
		var v = vidx - 12;
		var n = nidx - 9;

		for(i=0; i<4; i++) {
			if(gl_use_color) {
				gl_color[vidx] = gl_color[v];
			}
			if(gl_use_texcoord) {
				gl_texcoord[vidx] = gl_texcoord[v];
			}
			gl_vertex[vidx++] = gl_vertex[v++];

			if(gl_use_normal && i < 3) {
				gl_normal[nidx++] = gl_normal[n++];
			}
		}
		v += 4;
		n += 3;
		for(i=0; i<4; i++) {
			if(gl_use_color) {
				gl_color[vidx] = gl_color[v];
			}
			if(gl_use_texcoord) {
				gl_texcoord[vidx] = gl_texcoord[v];
			}
			gl_vertex[vidx++] = gl_vertex[v++];

			if(gl_use_normal && i < 3) {
				gl_normal[nidx++] = gl_normal[n++];
			}
		}

		gl_nverts += 2;
	}

	gl_vertex[vidx] = x;
	gl_vertex[vidx + 1] = y;
	gl_vertex[vidx + 2] = z;
	gl_vertex[vidx + 3] = 1.0;

	if(gl_use_color) {
		gl_color[vidx] = gl_curr_color[0];
		gl_color[vidx + 1] = gl_curr_color[1];
		gl_color[vidx + 2] = gl_curr_color[2];
		gl_color[vidx + 3] = gl_curr_color[3];
	}

	if(gl_use_normal) {
		gl_normal[nidx] = gl_curr_normal[0];
		gl_normal[nidx + 1] = gl_curr_normal[1];
		gl_normal[nidx + 2] = gl_curr_normal[2];
	}

	if(gl_use_texcoord) {
		gl_texcoord[vidx] = gl_curr_texcoord[0];
		gl_texcoord[vidx + 1] = gl_curr_texcoord[1];
		gl_texcoord[vidx + 2] = gl_texcoord[vidx + 3] = 0.0;
	}

	gl_vert_calls++;
	gl_nverts++;

	var buffer_full;
	if(gl_prim == GL_QUADS) {
		/* leave space for 6 more worst-case and don't allow flushes mid-quad */
		buffer_full = gl_nverts >= GL_MAX_VERTS - 6 && gl_vert_calls % 4 == 0;
	} else {
		buffer_full = gl_nverts >= GL_MAX_VERTS - gl_prim;
	}

	if(buffer_full) {
		gl_draw_immediate();
		glBegin(gl_prim);	/* reset everything */
	}
}

function glNormal3f(x, y, z)
{
	gl_curr_normal[0] = x;
	gl_curr_normal[1] = y;
	gl_curr_normal[3] = z;
}

function glColor3f(r, g, b)
{
	gl_curr_color[0] = r;
	gl_curr_color[1] = g;
	gl_curr_color[2] = b;
	gl_curr_color[3] = 1.0;
}

function glColor4f(r, g, b, a)
{
	gl_curr_color[0] = r;
	gl_curr_color[1] = g;
	gl_curr_color[2] = b;
	gl_curr_color[3] = a;
}

function glTexCoord1f(s)
{
	gl_curr_texcoord[0] = s;
	gl_curr_texcoord[1] = 0.0;
}

function glTexCoord2f(s, t)
{
	gl_curr_texcoord[0] = s;
	gl_curr_texcoord[1] = t;
}
